package kr.co.vividnext.sodalive.member

import com.querydsl.jpa.impl.JPAQueryFactory
import kr.co.vividnext.sodalive.fcm.GetMessageRecipientPushTokenResponse
import kr.co.vividnext.sodalive.fcm.QGetMessageRecipientPushTokenResponse
import kr.co.vividnext.sodalive.member.QMember.member
import kr.co.vividnext.sodalive.member.auth.QAuth.auth
import kr.co.vividnext.sodalive.member.block.BlockMemberRepository
import kr.co.vividnext.sodalive.member.following.QCreatorFollowing.creatorFollowing
import kr.co.vividnext.sodalive.member.notification.QMemberNotification.memberNotification
import kr.co.vividnext.sodalive.message.QMessage.message
import org.springframework.data.jpa.repository.JpaRepository
import org.springframework.stereotype.Repository

@Repository
interface MemberRepository : JpaRepository<Member, Long>, MemberQueryRepository {
    fun findByEmail(email: String): Member?
    fun findByNickname(nickname: String): Member?
}

interface MemberQueryRepository {
    fun findByPushToken(pushToken: String): List<Member>
    fun findByNicknameAndOtherCondition(nickname: String, memberId: Long): List<Member>
    fun findCreatorByIdOrNull(memberId: Long): Member?
    fun getAllRecipientPushTokens(isAuth: Boolean, container: String): List<List<String>>
    fun getCreateLiveRoomNotificationRecipientPushTokens(
        creatorId: Long,
        isAuth: Boolean,
        container: String
    ): List<List<String>>

    fun getUploadContentNotificationRecipientPushTokens(
        creatorId: Long,
        isAuth: Boolean,
        container: String
    ): List<List<String>>

    fun getMessageRecipientPushToken(messageId: Long): GetMessageRecipientPushTokenResponse
    fun getIndividualRecipientPushTokens(recipients: List<Long>, isAuth: Boolean): Map<String, List<List<String>>>
}

@Repository
class MemberQueryRepositoryImpl(
    private val queryFactory: JPAQueryFactory,
    private val blockMemberRepository: BlockMemberRepository
) : MemberQueryRepository {
    override fun findByPushToken(pushToken: String): List<Member> {
        return queryFactory
            .selectFrom(member)
            .where(member.pushToken.eq(pushToken))
            .fetch()
    }

    override fun findByNicknameAndOtherCondition(nickname: String, memberId: Long): List<Member> {
        return queryFactory
            .selectFrom(member)
            .where(
                member.nickname.containsIgnoreCase(nickname)
                    .and(member.id.ne(memberId))
                    .and(member.role.ne(MemberRole.ADMIN))
                    .and(member.role.ne(MemberRole.AGENT))
            )
            .fetch()
    }

    override fun findCreatorByIdOrNull(memberId: Long): Member? {
        return queryFactory
            .selectFrom(member)
            .where(
                member.id.eq(memberId)
                    .and(member.role.eq(MemberRole.CREATOR))
            )
            .fetchFirst()
    }

    override fun getAllRecipientPushTokens(isAuth: Boolean, container: String): List<List<String>> {
        var where = member.isActive.isTrue
            .and(member.email.notIn("admin@sodalive.net"))
            .and(member.container.eq(container))

        if (isAuth) {
            where = where.and(member.auth.isNotNull)
        }

        return queryFactory
            .select(member.pushToken)
            .from(member)
            .leftJoin(member.auth, auth)
            .where(where)
            .fetch()
            .toSet()
            .chunked(500)
    }

    override fun getCreateLiveRoomNotificationRecipientPushTokens(
        creatorId: Long,
        isAuth: Boolean,
        container: String
    ): List<List<String>> {
        val member = QMember.member
        val creator = QMember.member

        var where = creatorFollowing.isActive.isTrue
            .and(creator.id.eq(creatorId))
            .and(member.email.notIn("admin@sodalive.net"))
            .and(member.container.eq(container))
            .and(memberNotification.live.isTrue)
            .and(
                member.id.notIn(
                    blockMemberRepository.getBlockedMemberList(creatorId)
                )
            )

        if (isAuth) {
            where = where.and(member.auth.isNotNull)
        }

        return queryFactory
            .select(member.pushToken)
            .from(creatorFollowing)
            .innerJoin(creatorFollowing.creator, creator)
            .innerJoin(creatorFollowing.member, member)
            .innerJoin(member.notification, memberNotification)
            .leftJoin(member.auth, auth)
            .where(where)
            .fetch()
            .toSet()
            .chunked(500)
    }

    override fun getUploadContentNotificationRecipientPushTokens(
        creatorId: Long,
        isAuth: Boolean,
        container: String
    ): List<List<String>> {
        val member = QMember.member
        val creator = QMember.member

        var where = creatorFollowing.isActive.isTrue
            .and(creator.id.eq(creatorId))
            .and(member.email.notIn("admin@sodalive.net"))
            .and(member.container.eq(container))
            .and(memberNotification.uploadContent.isTrue)
            .and(
                member.id.notIn(
                    blockMemberRepository.getBlockedMemberList(creatorId)
                )
            )

        if (isAuth) {
            where = where.and(member.auth.isNotNull)
        }

        return queryFactory
            .select(member.pushToken)
            .from(creatorFollowing)
            .innerJoin(creatorFollowing.creator, creator)
            .innerJoin(creatorFollowing.member, member)
            .innerJoin(member.notification, memberNotification)
            .leftJoin(member.auth, auth)
            .where(where)
            .fetch()
            .toSet()
            .chunked(500)
    }

    override fun getMessageRecipientPushToken(messageId: Long): GetMessageRecipientPushTokenResponse {
        return queryFactory
            .select(
                QGetMessageRecipientPushTokenResponse(
                    member.pushToken,
                    member.container
                )
            )
            .from(message)
            .innerJoin(message.recipient, member)
            .where(message.id.eq(messageId))
            .fetchFirst()
    }

    override fun getIndividualRecipientPushTokens(
        recipients: List<Long>,
        isAuth: Boolean
    ): Map<String, List<List<String>>> {
        var where = member.isActive.isTrue
            .and(member.email.notIn("admin@sodalive.net"))

        if (isAuth) {
            where = where.and(member.auth.isNotNull)
        }

        val aosPushTokens = queryFactory
            .select(member.pushToken)
            .from(member)
            .leftJoin(member.auth, auth)
            .where(where.and(member.container.eq("aos")))
            .fetch()
            .toSet()
            .chunked(500)

        val iosPushTokens = queryFactory
            .select(member.pushToken)
            .from(member)
            .leftJoin(member.auth, auth)
            .where(where.and(member.container.eq("ios")))
            .fetch()
            .toSet()
            .chunked(500)

        return mapOf("aos" to aosPushTokens, "ios" to iosPushTokens)
    }
}
